from abc import abstractmethod
from functools import reduce
import itertools as it
from fractions import Fraction
import math
import numpy as np
from pathlib import Path
import pickle
import random
from typing import Optional

from .ballot import Ballot
from .pref_profile import PreferenceProfile


class BallotGenerator:
    """
    Base class for ballot generation models
    """

    def __init__(
        self,
        candidates: list,
        *,
        ballot_length: Optional[int] = None,
        pref_interval_by_bloc=None,
        bloc_voter_prop=None,
    ):

        """
         Initializes a Ballot Generator

        Args:
            candidates (list): list of candidates in the election
            ballot_length (Optional[int]): length of ballots to generate.
            Defaults to the length of candidates.
            pref_interval_by_bloc (dict[dict], optional): a mapping of slate to preference interval
            (ex. {race: {candidate : interval length}})
            bloc_voter_prop (dict): a mapping of slate to voter proportions
            (ex. {race: voter proportion}). Defaults to None.

        Raises:
            ValueError: if the voter proportion for blocs don't sum to 1
            ValueError: if preference interval for candidates must sum to 1
            ValueError: slates and blocs are not the same
        """

        self.ballot_length = (
            ballot_length if ballot_length is not None else len(candidates)
        )
        self.candidates = candidates

        if bloc_voter_prop and pref_interval_by_bloc:  # PL, BT, AC, CS
            if round(sum(bloc_voter_prop.values())) != 1:
                raise ValueError("Voter proportion for blocs must sum to 1")
            for interval in pref_interval_by_bloc.values():
                if round(sum(interval.values())) != 1:
                    raise ValueError("Preference interval for candidates must sum to 1")
            if bloc_voter_prop.keys() != pref_interval_by_bloc.keys():
                raise ValueError("slates and blocs are not the same")

            self.pref_interval_by_bloc = pref_interval_by_bloc
            self.bloc_voter_prop = bloc_voter_prop

    @classmethod
    def from_params(
        cls,
        slate_to_candidates: dict,
        bloc_voter_prop: dict,
        cohesion: dict,
        alphas: dict,
        **data,
    ):
        """
        Initializes a Ballot Generator by constructing a preference interval from parameters
        (the prior parameters will be overwrittern)

        Args:
            slate_to_candidate (dict): a mapping of slate to candidates
            (ex. {race: [candidate]})
            bloc_voter_prop (dict): a mapping of the percentage of total voters per bloc
            (ex. {race: 0.5})
            cohesion (dict): cohension factor for each bloc
            alphas (dict): alpha for the dirchlet distribution of each bloc

        Raises:
            ValueError: if the voter proportion for blocs don't sum to 1
            ValueError: slates and blocs are not the same

        Returns:
            BallotGenerator: initialized ballot generator
        """

        if sum(bloc_voter_prop.values()) != 1.0:
            raise ValueError(
                f"bloc proportions ({bloc_voter_prop.values()}) do not equal 1"
            )
        if slate_to_candidates.keys() != bloc_voter_prop.keys():
            raise ValueError("slates and blocs are not the same")

        def _construct_preference_interval(
            alphas: dict, cohesion: int, bloc: str, slate_to_cands: dict
        ) -> dict:
            intervals = {}

            for group, alpha in alphas.items():
                num_cands = len(slate_to_cands[group])
                alpha = [alpha] * num_cands
                probs = list(np.random.default_rng().dirichlet(alpha=alpha))
                for prob, cand in zip(probs, slate_to_cands[group]):
                    if group == bloc:  # e.g W for W cands
                        pi = cohesion
                    else:  # e.g W for POC cands
                        pi = 1 - cohesion
                    intervals[cand] = pi * prob

            return intervals

        interval_by_bloc = {}
        for bloc in bloc_voter_prop:
            interval = _construct_preference_interval(
                alphas[bloc], cohesion[bloc], bloc, slate_to_candidates
            )
            interval_by_bloc[bloc] = interval

        if "candidates" not in data:
            cands = list(
                {cand for cands in slate_to_candidates.values() for cand in cands}
            )
            data["candidates"] = cands

        if "pref_interval_by_bloc" not in data:
            data["pref_interval_by_bloc"] = interval_by_bloc

        if "bloc_voter_prop" not in data:
            data["bloc_voter_prop"] = bloc_voter_prop

        generator = cls(**data)

        if isinstance(generator, (AlternatingCrossover, CambridgeSampler)):
            generator.slate_to_candidates = slate_to_candidates

        return generator

    @abstractmethod
    def generate_profile(self, number_of_ballots: int) -> PreferenceProfile:
        """
        Generates a preference profile

        Args:
            number_of_ballots (int): number of ballots to generate
        Returns:
            PreferenceProfile: a generated preference profiles
        """
        pass

    @staticmethod
    def round_num(num: float) -> int:
        """
        rounds up or down a float randomly

        Args:
            num (float): number to round

        Returns:
            int: a whole number
        """
        rand = np.random.random()
        return math.ceil(num) if rand > 0.5 else math.floor(num)

    @staticmethod
    def ballot_pool_to_profile(ballot_pool, candidates) -> PreferenceProfile:
        """
        Given a list of ballots and candidates, convert them into a Preference Profile

        Args:
            ballot_pool (list of tuple): a list of ballots, with tuple as their ranking
            candidates (list): a list of candidates

        Returns:
            PreferenceProfile: a preference profile representing the ballots in the election
        """
        ranking_counts: dict[tuple, int] = {}
        ballot_list: list[Ballot] = []

        for ranking in ballot_pool:
            tuple_rank = tuple(ranking)
            ranking_counts[tuple_rank] = (
                ranking_counts[tuple_rank] + 1 if tuple_rank in ranking_counts else 1
            )

        for ranking, count in ranking_counts.items():
            rank = [set([cand]) for cand in ranking]
            b = Ballot(ranking=rank, weight=Fraction(count))
            ballot_list.append(b)

        return PreferenceProfile(ballots=ballot_list, candidates=candidates)


class BallotSimplex(BallotGenerator):
    """
    Base class for ballot generation models
    """

    def __init__(
        self, alpha: Optional[float] = None, point: Optional[dict] = None, **data
    ):
        """
        Initializes a Ballot Simplex model

        Args:
            alpha (float, optional): alpha parameter for ballot simplex. Defaults to None.
            point (dict, optional): a point in the ballot simplex,
            with candidate as keys and electoral support as values. Defaults to None.

        Raises:
            AttributeError: if point and alpha are not initialized
        """
        if alpha is None and point is None:
            raise AttributeError("point or alpha must be initialized")
        self.alpha = alpha
        if alpha == float("inf"):
            self.alpha = 1e20
        if alpha == 0:
            self.alpha = 1e-10
        self.point = point
        super().__init__(**data)

    @classmethod
    def from_point(cls, point: dict, **data):
        """
        Initializes a Ballot Simplex model from a point in the dirichlet distribution

        Args:
            point (dict): a mapping of candidate to candidate support

        Raises:
            ValueError: if the candidate support does not sum to 1

        Returns:
            BallotSimplex: initialized from point
        """
        if sum(point.values()) != 1.0:
            raise ValueError(
                f"probability distribution from point ({point.values()}) does not sum to 1"
            )
        return cls(point=point, **data)

    @classmethod
    def from_alpha(cls, alpha: float, **data):
        """
        Initializes a Ballot Simplex model from an alpha value for the dirichlet distribution

        Args:
            alpha (float): an alpha parameter for the dirichlet distribution

        Returns:
            BallotSimplex: initialized from alpha
        """

        return cls(alpha=alpha, **data)

    def generate_profile(self, number_of_ballots) -> PreferenceProfile:
        perm_set = it.permutations(self.candidates, self.ballot_length)

        perm_rankings = [list(value) for value in perm_set]

        if self.alpha is not None:
            draw_probabilities = list(
                np.random.default_rng().dirichlet([self.alpha] * len(perm_rankings))
            )
        elif self.point:
            # calculates probabilities for each ranking
            # using probability distribution for candidate support
            draw_probabilities = [
                reduce(
                    lambda prod, cand: prod * self.point[cand] if self.point else 0,
                    ranking,
                    1.0,
                )
                for ranking in perm_rankings
            ]
            draw_probabilities = [
                prob / sum(draw_probabilities) for prob in draw_probabilities
            ]

        ballot_pool = []

        for _ in range(number_of_ballots):
            index = np.random.choice(
                range(len(perm_rankings)), 1, p=draw_probabilities
            )[0]
            ballot_pool.append(perm_rankings[index])

        return self.ballot_pool_to_profile(ballot_pool, self.candidates)


class ImpartialCulture(BallotSimplex):
    """
    Impartial Culture model (child class of BallotSimplex)
    with an alpha value of 1e10 (should be infinity theoretically)
    """

    def __init__(self, **data):
        super().__init__(alpha=float("inf"), **data)


class ImpartialAnonymousCulture(BallotSimplex):
    """
    Impartial Anonymous Culture model (child class of BallotSimplex)
    with an alpha value of 1
    """

    def __init__(self, **data):
        super().__init__(alpha=1, **data)


class PlackettLuce(BallotGenerator):
    """
    Plackett Luce Ballot Generation Model (child class of BallotGenerator)
    """

    def __init__(self, **data):
        """
        Initializes Plackett Luce Ballot Generation Model

        Args:
            pref_interval_by_bloc (dict): a mapping of slate to preference interval
            (ex. {race: {candidate : interval length}})
            bloc_voter_prop (dict): a mapping of slate to voter proportions
            (ex. {race: voter proportion})
        """

        # Call the parent class's __init__ method to handle common parameters
        super().__init__(**data)

    def generate_profile(self, number_of_ballots) -> PreferenceProfile:
        ballot_pool = []

        for bloc in self.bloc_voter_prop.keys():
            # number of voters in this bloc
            num_ballots = self.round_num(number_of_ballots * self.bloc_voter_prop[bloc])
            pref_interval_dict = self.pref_interval_by_bloc[bloc]
            # creates the interval of probabilities for candidates supported by this block
            cand_support_vec = [pref_interval_dict[cand] for cand in self.candidates]

            for _ in range(num_ballots):
                # generates ranking based on probability distribution of candidate support
                ballot = list(
                    np.random.choice(
                        self.candidates,
                        self.ballot_length,
                        p=cand_support_vec,
                        replace=False,
                    )
                )

                ballot_pool.append(ballot)

        pp = self.ballot_pool_to_profile(
            ballot_pool=ballot_pool, candidates=self.candidates
        )
        return pp


class BradleyTerry(BallotGenerator):
    """
    Bradley Terry Ballot Generation Model (child class of BallotGenerator)
    """

    def __init__(self, **data):
        """
        Initializes a Bradley Terry Ballot Generation Model

        Args:
            pref_interval_by_bloc (dict): a mapping of slate to preference interval
            (ex. {race: {candidate : interval length}})
            bloc_voter_prop (dict): a mapping of slate to voter proportions
            (ex. {race: voter proportion})
        """

        # Call the parent class's __init__ method to handle common parameters
        super().__init__(**data)

    def _calc_prob(self, permutations: list[tuple], cand_support_dict: dict) -> dict:
        """
        given a list of rankings and the preference interval,
        calculates the probability of observing each ranking

        Args:
            permutations (list[tuple]): a list of permuted rankings
            cand_support_dict (dict): a mapping from candidate to their
            support (preference interval)

        Returns:
            dict: a mapping of the rankings to their probability
        """
        ranking_to_prob = {}
        for ranking in permutations:
            prob = 1
            for i in range(len(ranking)):
                cand_i = ranking[i]
                greater_cand_support = cand_support_dict[cand_i]
                for j in range(i + 1, len(ranking)):
                    cand_j = ranking[j]
                    cand_support = cand_support_dict[cand_j]
                    prob *= greater_cand_support / (greater_cand_support + cand_support)
            ranking_to_prob[ranking] = prob
        return ranking_to_prob

    def generate_profile(self, number_of_ballots) -> PreferenceProfile:

        permutations = list(it.permutations(self.candidates, self.ballot_length))
        ballot_pool: list[list] = []

        for bloc in self.bloc_voter_prop.keys():
            num_ballots = self.round_num(number_of_ballots * self.bloc_voter_prop[bloc])
            pref_interval_dict = self.pref_interval_by_bloc[bloc]

            ranking_to_prob = self._calc_prob(
                permutations=permutations, cand_support_dict=pref_interval_dict
            )

            indices = range(len(ranking_to_prob))
            prob_distrib = list(ranking_to_prob.values())
            prob_distrib = [float(p) / sum(prob_distrib) for p in prob_distrib]

            ballots_indices = np.random.choice(
                indices,
                num_ballots,
                p=prob_distrib,
                replace=True,
            )

            rankings = list(ranking_to_prob.keys())
            ballots = [rankings[i] for i in ballots_indices]

            ballot_pool = ballot_pool + ballots

        pp = self.ballot_pool_to_profile(
            ballot_pool=ballot_pool, candidates=self.candidates
        )
        return pp


class AlternatingCrossover(BallotGenerator):
    """
    Alternating Crossover Ballot Generation Model (child class of BallotGenerator)
    """

    def __init__(
        self,
        slate_to_candidates=None,
        bloc_crossover_rate=None,
        **data,
    ):
        """
        Initializes Alternating Crossover Ballot Generation Model

        Args:
            slate_to_candidate (dict): a mapping of slate to candidates
            (ex. {race: [candidate]})
            pref_interval_by_bloc (dict): a mapping of bloc to preference interval
            (ex. {race: {candidate : interval length}})
            bloc_voter_prop (dict): a mapping of the percentage of total voters per bloc
            (ex. {race: 0.5})
            bloc_crossover_rate (dict): a mapping of percentage of crossover voters per bloc
            (ex. {race: {other_race: 0.5}})
        """
        # Call the parent class's __init__ method to handle common parameters
        super().__init__(**data)

        self.slate_to_candidates = slate_to_candidates
        self.bloc_crossover_rate = bloc_crossover_rate

    def generate_profile(self, number_of_ballots) -> PreferenceProfile:

        ballot_pool = []

        for bloc in self.bloc_voter_prop.keys():

            num_ballots = self.round_num(number_of_ballots * self.bloc_voter_prop[bloc])
            crossover_dict = self.bloc_crossover_rate[bloc]
            pref_interval_dict = self.pref_interval_by_bloc[bloc]

            # generates crossover ballots from each bloc (allowing for more than two blocs)
            for opposing_slate in crossover_dict.keys():
                crossover_rate = crossover_dict[opposing_slate]
                num_crossover_ballots = self.round_num(crossover_rate * num_ballots)

                opposing_cands = self.slate_to_candidates[opposing_slate]
                bloc_cands = self.slate_to_candidates[bloc]

                for _ in range(num_crossover_ballots):
                    pref_for_opposing = [
                        pref_interval_dict[cand] for cand in opposing_cands
                    ]
                    # convert to probability distribution
                    pref_for_opposing = [
                        p / sum(pref_for_opposing) for p in pref_for_opposing
                    ]

                    pref_for_bloc = [pref_interval_dict[cand] for cand in bloc_cands]
                    # convert to probability distribution
                    pref_for_bloc = [p / sum(pref_for_bloc) for p in pref_for_bloc]

                    bloc_cands = list(
                        np.random.choice(
                            bloc_cands,
                            p=pref_for_bloc,
                            size=len(bloc_cands),
                            replace=False,
                        )
                    )
                    opposing_cands = list(
                        np.random.choice(
                            opposing_cands,
                            size=len(opposing_cands),
                            p=pref_for_opposing,
                            replace=False,
                        )
                    )

                    # alternate the bloc and opposing bloc candidates to create crossover ballots
                    if bloc != opposing_slate:  # alternate
                        ballot = [
                            item
                            for pair in zip(opposing_cands, bloc_cands)
                            for item in pair
                            if item is not None
                        ]

                    # check that ballot_length is shorter than total number of cands
                    ballot_pool.append(ballot)

                # Bloc ballots
                for _ in range(num_ballots - num_crossover_ballots):
                    ballot = bloc_cands + opposing_cands
                    ballot_pool.append(ballot)

        pp = self.ballot_pool_to_profile(
            ballot_pool=ballot_pool, candidates=self.candidates
        )
        return pp


class OneDimSpatial(BallotGenerator):
    def generate_profile(self, number_of_ballots) -> PreferenceProfile:
        candidate_position_dict = {c: np.random.normal(0, 1) for c in self.candidates}
        voter_positions = np.random.normal(0, 1, number_of_ballots)

        ballot_pool = []

        for vp in voter_positions:
            distance_dict = {
                c: abs(v - vp) for c, v, in candidate_position_dict.items()
            }
            candidate_order = sorted(distance_dict, key=distance_dict.__getitem__)
            ballot_pool.append(candidate_order)

        return self.ballot_pool_to_profile(ballot_pool, self.candidates)


class CambridgeSampler(BallotGenerator):
    """
    Cambridge Sampler Ballot Generation  model (child class of BallotGenerator)
    """

    def __init__(
        self,
        slate_to_candidates=None,
        bloc_crossover_rate=None,
        path: Optional[Path] = None,
        **data,
    ):
        """
        Initializes Cambridge Sampler Ballot Generation  model

        Args:
            slate_to_candidate (dict): a mapping of slate to candidates
            (ex. {race: [candidate]})
            pref_interval_by_bloc (dict): a mapping of bloc to preference interval
            (ex. {race: {candidate : interval length}})
            path (Optional[Path]): a path to an election data file to sample from.
            Defaults to Cambridge elections.
        """

        # Call the parent class's __init__ method to handle common parameters
        super().__init__(**data)

        self.slate_to_candidates = slate_to_candidates
        self.bloc_crossover_rate = bloc_crossover_rate

        if path:
            self.path = path
        else:
            BASE_DIR = Path(__file__).resolve().parent
            DATA_DIR = BASE_DIR / "data/"
            self.path = Path(DATA_DIR, "Cambridge_09to17_ballot_types.p")

    def generate_profile(self, number_of_ballots: int) -> PreferenceProfile:

        with open(self.path, "rb") as pickle_file:
            ballot_frequencies = pickle.load(pickle_file)

        ballot_pool = []

        blocs = self.slate_to_candidates.keys()
        for bloc in blocs:
            # compute the number of voters in this bloc
            bloc_voters = self.round_num(self.bloc_voter_prop[bloc] * number_of_ballots)

            # store the opposition bloc
            opp_bloc = next(iter(set(blocs).difference(set(bloc))))

            # compute how many ballots list a bloc candidate first
            bloc_first_count = sum(
                [
                    freq
                    for ballot, freq in ballot_frequencies.items()
                    if ballot[0] == bloc
                ]
            )

            # Compute the pref interval for this bloc
            pref_interval_dict = self.pref_interval_by_bloc[bloc]

            # compute the relative probabilities of each ballot
            # sorted by ones where the ballot lists the bloc first
            # and those that list the opp first
            prob_ballot_given_bloc_first = {
                ballot: freq / bloc_first_count
                for ballot, freq in ballot_frequencies.items()
                if ballot[0] == bloc
            }
            prob_ballot_given_opp_first = {
                ballot: freq / bloc_first_count
                for ballot, freq in ballot_frequencies.items()
                if ballot[0] == opp_bloc
            }

            # Generate ballots
            for _ in range(bloc_voters):
                # Randomly choose first choice based off
                # bloc crossover rate
                first_choice = np.random.choice(
                    [bloc, opp_bloc],
                    p=[
                        1 - self.bloc_crossover_rate[bloc][opp_bloc],
                        self.bloc_crossover_rate[bloc][opp_bloc],
                    ],
                )
                # Based on first choice, randomly choose
                # ballots weighted by Cambridge frequency
                if first_choice == bloc:
                    bloc_ordering = random.choices(
                        list(prob_ballot_given_bloc_first.keys()),
                        weights=list(prob_ballot_given_bloc_first.values()),
                        k=1,
                    )[0]
                else:
                    bloc_ordering = random.choices(
                        list(prob_ballot_given_opp_first.keys()),
                        weights=list(prob_ballot_given_opp_first.values()),
                        k=1,
                    )[0]

                # Now turn bloc ordering into candidate ordering
                pl_ordering = list(
                    np.random.choice(
                        list(pref_interval_dict.keys()),
                        self.ballot_length,
                        p=list(pref_interval_dict.values()),
                        replace=False,
                    )
                )
                ordered_bloc_slate = [
                    c for c in pl_ordering if c in self.slate_to_candidates[bloc]
                ]
                ordered_opp_slate = [
                    c for c in pl_ordering if c in self.slate_to_candidates[opp_bloc]
                ]

                # Fill in the bloc slots as determined
                # With the candidate ordering generated with PL
                full_ballot = []
                for b in bloc_ordering:
                    if b == bloc:
                        if ordered_bloc_slate:
                            full_ballot.append(ordered_bloc_slate.pop(0))
                    else:
                        if ordered_opp_slate:
                            full_ballot.append(ordered_opp_slate.pop(0))

                ballot_pool.append(tuple(full_ballot))

        pp = self.ballot_pool_to_profile(
            ballot_pool=ballot_pool, candidates=self.candidates
        )
        return pp



from collections import namedtuple
from fractions import Fraction
import numpy as np
import random
from typing import Union, Iterable, Optional

from .ballot import Ballot
from .pref_profile import PreferenceProfile


COLOR_LIST = [
    (0.55, 0.71, 0.0),
    (0.82, 0.1, 0.26),
    (0.44, 0.5, 0.56),
    (1.0, 0.75, 0.0),
    (1.0, 0.77, 0.05),
    (0.0, 0.42, 0.24),
    (0.13, 0.55, 0.13),
    (0.9, 0.13, 0.13),
    (0.08, 0.38, 0.74),
    (0.41, 0.21, 0.61),
    (1.0, 0.72, 0.77),
    (1.0, 0.66, 0.07),
    (1.0, 0.88, 0.21),
    (0.55, 0.82, 0.77),
]

# Election Helper Functions
CandidateVotes = namedtuple("CandidateVotes", ["cand", "votes"])


def compute_votes(candidates: list, ballots: list[Ballot]) -> list[CandidateVotes]:
    """
    Computes first place votes for all candidates in a preference profile

    Args:
        candidates: List of all candidates in a PreferenceProfile
        ballots: List of Ballot objects

    Returns:
        List of tuples (candidate, number of votes) ordered by first place votes
    """
    votes = {cand: Fraction(0) for cand in candidates}

    for ballot in ballots:
        if not ballot.ranking:
            continue
        first_place_cand = unset(ballot.ranking[0])
        if isinstance(first_place_cand, list):
            for cand in first_place_cand:
                votes[cand] += ballot.weight / len(first_place_cand)
        else:
            votes[first_place_cand] += ballot.weight

    ordered = [
        CandidateVotes(cand=key, votes=value)
        for key, value in sorted(votes.items(), key=lambda x: x[1], reverse=True)
    ]

    return ordered


def fractional_transfer(
    winner: str, ballots: list[Ballot], votes: dict, threshold: int
) -> list[Ballot]:
    """
    Calculates fractional transfer from winner, then removes winner
    from the list of ballots

    Args:
        winner: Candidate to transfer votes from
        ballots: List of Ballot objects
        votes: Contains candidates and their corresponding vote totals
        threshold: Value required to be elected, used to calculate transfer value

    Returns:
        Modified ballots with transfered weights and the winning canidated removed
    """
    transfer_value = (votes[winner] - threshold) / votes[winner]

    for ballot in ballots:
        new_ranking = []
        if ballot.ranking and ballot.ranking[0] == {winner}:
            ballot.weight = ballot.weight * transfer_value
            for cand in ballot.ranking:
                if cand != {winner}:
                    new_ranking.append(cand)

    return remove_cand(winner, ballots)


def random_transfer(
    winner: str, ballots: list[Ballot], votes: dict, threshold: int
) -> list[Ballot]:
    """
    Cambridge-style transfer where transfer ballots are selected randomly

    Args:
        winner: Candidate to transfer votes from
        ballots: List of Ballot objects
        votes: Contains candidates and their corresponding vote totals
        threshold: Value required to be elected, used to calculate transfer value

    Returns:
        Modified ballots with transfered weights and the winning canidated removed
    """

    # turn all of winner's ballots into (multiple) ballots of weight 1
    weight_1_ballots = []
    for ballot in ballots:
        if ballot.ranking and ballot.ranking[0] == {winner}:
            # note: under random transfer, weights should always be integers
            for _ in range(int(ballot.weight)):
                weight_1_ballots.append(
                    Ballot(
                        id=ballot.id,
                        ranking=ballot.ranking,
                        weight=Fraction(1),
                        voters=ballot.voters,
                    )
                )

    # remove winner's ballots
    ballots = [
        ballot
        for ballot in ballots
        if not (ballot.ranking and ballot.ranking[0] == {winner})
    ]

    surplus_ballots = random.sample(weight_1_ballots, int(votes[winner]) - threshold)
    ballots += surplus_ballots

    transfered = remove_cand(winner, ballots)

    return transfered


def seqRCV_transfer(
    winner: str, ballots: list[Ballot], votes: dict, threshold: int
) -> list[Ballot]:
    """
    Transfer method Sequential RCV elections

    Args:
        winner: Candidate to transfer votes from
        ballots: List of Ballot objects
        votes: Contains candidates and their corresponding vote totals
        threshold: Value required to be elected, used to calculate transfer value

    Returns:
        Original list of ballots as Sequential RCV does not transfer votes
    """
    return ballots


def remove_cand(removed: Union[str, Iterable], ballots: list[Ballot]) -> list[Ballot]:
    """
    Removes specified candidate(s) from ballots

    Args:
        removed: Candidate or set of candidates to be removed
        ballots: List of Ballots to remove canidate(s) from

    Returns:
        Updated list of ballots with candidate(s) removed
    """

    if isinstance(removed, str):
        remove_set = {removed}
    elif isinstance(removed, Iterable):
        remove_set = set(removed)

    update = []
    for ballot in ballots:
        new_ranking = []
        if len(remove_set) == 1 and remove_set in ballot.ranking:
            for s in ballot.ranking:
                new_s = s.difference(remove_set)
                if new_s:
                    new_ranking.append(new_s)
            update.append(
                Ballot(
                    id=ballot.id,
                    ranking=new_ranking,
                    weight=ballot.weight,
                    voters=ballot.voters,
                )
            )
        elif len(remove_set) > 1:
            for s in ballot.ranking:
                new_s = s.difference(remove_set)
                if new_s:
                    new_ranking.append(new_s)
            update.append(
                Ballot(
                    id=ballot.id,
                    ranking=new_ranking,
                    weight=ballot.weight,
                    voters=ballot.voters,
                )
            )
        else:
            update.append(ballot)

    return update


# Summmary Stat functions
def first_place_votes(profile: PreferenceProfile) -> dict:
    """
    Calculates first-place votes for a PreferenceProfile

    Args:
        profile: Inputed profile of ballots

    Returns:
        Dictionary of candidates (keys) and first place vote totals (values)
    """
    cands = profile.get_candidates()
    ballots = profile.get_ballots()

    return {cand: float(votes) for cand, votes in compute_votes(cands, ballots)}


def mentions(profile: PreferenceProfile) -> dict:
    """
    Calculates total mentions for a PreferenceProfile

    Args:
        profile: Inputed profile of ballots

    Returns:
        Dictionary of candidates (keys) and mention totals (values)
    """
    mentions: dict[str, float] = {}

    ballots = profile.get_ballots()
    for ballot in ballots:
        for rank in ballot.ranking:
            for cand in rank:
                if cand not in mentions:
                    mentions[cand] = 0
                if len(rank) > 1:
                    mentions[cand] += (1 / len(rank)) * int(
                        ballot.weight
                    )  # split mentions for candidates that are tied
                else:
                    mentions[cand] += float(ballot.weight)

    return mentions


def borda_scores(
    profile: PreferenceProfile,
    ballot_length: Optional[int] = None,
    score_vector: Optional[list] = None,
) -> dict:
    """
    Calculates Borda scores for a PreferenceProfile

    Args:
        profile: Inputed profile of ballots
        ballot_length: Length of a ballot, if None length of longest ballot is \n
        is used
        score_vector: Borda weights, if None assigned based length of the \n
        longest ballot

    Returns:
        Dictionary of candidates (keys) and Borda scores (values)
    """
    candidates = profile.get_candidates()
    if ballot_length is None:
        ballot_length = max([len(ballot.ranking) for ballot in profile.ballots])
    if score_vector is None:
        score_vector = list(range(ballot_length, 0, -1))

    candidate_borda = {c: Fraction(0) for c in candidates}
    for ballot in profile.ballots:
        current_ind = 0
        candidates_covered = []
        for s in ballot.ranking:
            position_size = len(s)
            local_score_vector = score_vector[current_ind : current_ind + position_size]
            borda_allocation = sum(local_score_vector) / position_size
            for c in s:
                candidate_borda[c] += Fraction(borda_allocation) * ballot.weight
            current_ind += position_size
            candidates_covered += list(s)

        # If ballot was incomplete, evenly allocation remaining points
        if current_ind < len(score_vector):
            remainder_cands = set(candidates).difference(set(candidates_covered))
            remainder_score_vector = score_vector[current_ind:]
            remainder_borda_allocation = sum(remainder_score_vector) / len(
                remainder_cands
            )
            for c in remainder_cands:
                candidate_borda[c] += (
                    Fraction(remainder_borda_allocation) * ballot.weight
                )

    return candidate_borda


def unset(input: set):
    """
    Removes object from set

    Args:
        input: Input set

    Returns:
        If set has length one returns the object, else returns a list
    """
    rv = list(input)

    if len(rv) == 1:
        return rv[0]

    return rv


def candidate_position_dict(ranking: list[set[str]]) -> dict:
    """
    Creates a dictionary with the integer ranking of candidates given a set ranking \n
    i.e. A > B, C > D returns {A: 1, B: 2, C: 2, D: 4}

    Args:
        ranking: A list-of-set ranking of candidates

    Returns:
        Dictionary of candidates (keys) and integer rankings (values)
    """
    candidate_positions = {}
    position = 0

    for tie_set in ranking:
        for candidate in tie_set:
            candidate_positions[candidate] = position
        position += len(tie_set)

    return candidate_positions


def tie_broken_ranking(
    ranking: list[set[str]], profile: PreferenceProfile, tiebreak: str = "none"
) -> list[set[str]]:
    """
    Breaks ties in a list-of-sets ranking according to a given scheme

    Args:
        ranking: A list-of-set ranking of candidates
        profile: The election ballot profile
        tiebreak: Method of tiebreak, currently supports 'none', 'random', 'borda', 'firstplace'

    Returns:
        A list-of-set ranking of candidates (tie broken down to one candidate sets unless \n
        tiebreak = 'none')
    """

    new_ranking = []
    if tiebreak == "none":
        new_ranking = ranking
    elif tiebreak == "random":
        for s in ranking:
            shuffled_s = list(np.random.permutation(list(s)))
            new_ranking += [{c} for c in shuffled_s]
    elif tiebreak == "firstplace":
        tiebreak_scores = first_place_votes(profile)
        for s in ranking:
            ordered_set = scores_into_set_list(tiebreak_scores, s)
            new_ranking += ordered_set
    elif tiebreak == "borda":
        tiebreak_scores = borda_scores(profile)
        for s in ranking:
            ordered_set = scores_into_set_list(tiebreak_scores, s)
            new_ranking += ordered_set
    else:
        raise ValueError("Invalid tiebreak code was provided")

    if tiebreak != "none" and any(len(s) > 1 for s in new_ranking):
        print("Initial tiebreak was unsuccessful, performing random tiebreak")
        new_ranking = tie_broken_ranking(
            ranking=new_ranking, profile=profile, tiebreak="random"
        )

    return new_ranking


def scores_into_set_list(
    score_dict: dict, candidate_subset: Union[list[str], set[str], None] = None
) -> list[set[str]]:
    """
    Sorts candidates based on a scoring dictionary (i.e Borda, First-Place)

    Args:
        score_dict: Dictionary between candidates (key) and their score (value)
        candidate_subset: Relevant candidates to sort

    Returns:
        Candidate rankings in a list-of-sets form
    """
    if isinstance(candidate_subset, list):
        candidate_subset = set(candidate_subset)

    tier_dict: dict = {}
    for k, v in score_dict.items():
        if v in tier_dict.keys():
            tier_dict[v].add(k)
        else:
            tier_dict[v] = {k}
    tier_list = [tier_dict[k] for k in sorted(tier_dict.keys(), reverse=True)]
    if candidate_subset is not None:
        tier_list = [
            t & candidate_subset for t in tier_list if len(t & candidate_subset) > 0
        ]
    return tier_list


def elect_cands_from_set_ranking(
    ranking: list[set[str]], seats: int
) -> tuple[list[set[str]], list[set[str]]]:
    """
    Splits a ranking into elected and eliminated based on seats,
    and if a tie set overlaps the desired number of seats raises a ValueError

    Args:
        ranking: A list-of-set ranking of candidates
        seats: Number of seats to fill

    Returns:
        A list-of-sets of elected candidates, a list-of-sets of eliminated candidates
    """
    cands_elected = 0
    elected = []
    eliminated = []

    for i, s in enumerate(ranking):
        if cands_elected + len(s) <= seats:
            cands_elected += len(s)
            elected.append(s)
        else:
            eliminated = ranking[i:]
            break

    if cands_elected != seats:
        raise ValueError(
            "Cannot elect correct number of candidates without breaking ties."
        )

    return elected, eliminated



from fractions import Fraction
import os
import pandas as pd
from pandas.errors import EmptyDataError, DataError
import pathlib
from typing import Optional

from .pref_profile import PreferenceProfile
from .ballot import Ballot

# TODO: update documentation for function below


def rank_column_csv(
    fpath: str,
    *,
    weight_col: Optional[int] = None,
    delimiter: Optional[str] = None,
    id_col: Optional[int] = None,
) -> PreferenceProfile:
    """
    Takes a file path and loads a cast vote record (cvr) with cast votes as cols and voters as rows.
    Empty cells are treated as None. Currently, missing voter ids are not assigned

    :param fpath: :class:`str`: path to cvr file
    :param weight_col: optional :class:`int` the column position for ballot weights
    if parsing Scottish elections like cvrs
    :param delimiter: :class:`str`: the character that breaks up rows
    :param id_col: optional :class:`int`: index for the column with voter ids \n

    Raises:
        `FileNotFoundError`: if fpath is invalid \n
        `EmptyDataError`: if dataset is empty \n
        `ValueError`: if the voter id column has missing values \n
        `DataError`: if the voter id column has duplicate values

    :return: A preference profile object with all cast ballots from file
    :rtype: :class:`PreferenceProfile`
    """
    if not os.path.isfile(fpath):
        raise FileNotFoundError(f"File with path {fpath} cannot be found")

    cvr_path = pathlib.Path(fpath)
    df = pd.read_csv(
        cvr_path,
        on_bad_lines="error",
        encoding="utf8",
        index_col=False,
        delimiter=delimiter,
    )

    if df.empty:
        raise EmptyDataError("Dataset cannot be empty")
    if id_col is not None and df.iloc[:, id_col].isnull().values.any():  # type: ignore
        raise ValueError(f"Missing value(s) in column at index {id_col}")
    if id_col is not None and not df.iloc[:, id_col].is_unique:
        raise DataError(f"Duplicate value(s) in column at index {id_col}")

    ranks = list(df.columns)
    if id_col is not None:
        ranks.remove(df.columns[id_col])
    grouped = df.groupby(ranks, dropna=False)
    ballots = []

    for group, group_df in grouped:
        ranking = [{None} if pd.isnull(c) else {c} for c in group]
        voters = None
        if id_col is not None:
            voters = set(group_df.iloc[:, id_col])
        weight = len(group_df)
        if weight_col is not None:
            weight = sum(group_df.iloc[:, weight_col])
        b = Ballot(ranking=ranking, weight=Fraction(weight), voters=voters)
        ballots.append(b)

    return PreferenceProfile(ballots=ballots)


def blt(fpath: str) -> tuple[PreferenceProfile, int]:
    """
    Loads cast vote record from .blt file.
    blt is text-like format used for scottish election data
    the first line of the file is metadata recording the number of candidates and seats,
    followed by ballot data (first number in row is ballot weight),
    followed by candidate data (order corresponds to number in ballots),
    followed by election location
    :param fpath: :class:`str`: path to cvr file \n
    Raises:\n
    `FileNotFoundError`: if fpath is invalid \n
    `EmptyDataError`: if dataset is empty \n
    `DataError`: if there is missing or incorrect metadata or candidate data \n
    :return: preference profile with all cast ballots from file and number of seats in election
    :rtype: :class:`PreferenceProfile`  , :class:`int`

    """
    ballots = []
    names = []
    name_map = {}
    numbers = True
    cands_included = False

    if not os.path.isfile(fpath):
        raise FileNotFoundError(f"File with path {fpath} cannot be found")
    if os.path.getsize(fpath) == 0:
        raise EmptyDataError("Dataset cannot be empty")

    with open(fpath, "r") as file:
        for i, line in enumerate(file):
            s = line.rstrip("\n").rstrip()
            if i == 0:
                # first number is number of candidates, second is number of seats to elect
                metadata = [int(data) for data in s.split(" ")]
                if len(metadata) != 2:
                    raise DataError(
                        "metadata (first line) should have two parameters"
                        " (number of candidates, number of seats)"
                    )
                seats = metadata[1]
            # read in ballots, cleaning out rankings labeled '0' (designating end of line)
            elif numbers:
                ballot = [int(vote) for vote in s.split(" ")]
                num_votes = ballot[0]
                # ballots terminate with a single row with the character '0'
                if num_votes == 0:
                    numbers = False
                else:
                    ranking = [rank for rank in list(ballot[1:]) if rank != 0]
                    b = (ranking, num_votes)
                    ballots.append(b)  # this is converted to the PP format later
            # read in candidates
            elif "(" in s:
                cands_included = True
                name_parts = s.strip('"').split(" ")
                first_name = " ".join(name_parts[:-2])
                last_name = name_parts[-2]
                party = name_parts[-1].strip("(").strip(")")
                names.append(str((first_name, last_name, party)))
            else:
                if len(names) != metadata[0]:
                    err_message = (
                        f"Number of candidates listed, {len(names)}," + f" differs from"
                        f"number of candidates recorded in metadata, {metadata[0]}"
                    )
                    raise DataError(err_message)
                # read in election location (do we need this?)
                # location = s.strip("\"")
                if not cands_included:
                    raise DataError("Candidates missing from file")
                # map candidate numbers onto their names and convert ballots to PP format
                for i, name in enumerate(names):
                    name_map[i + 1] = name
                clean_ballots = [
                    Ballot(
                        ranking=[{name_map[cand]} for cand in ballot[0]],
                        weight=Fraction(ballot[1]),
                    )
                    for ballot in ballots
                ]

        return PreferenceProfile(ballots=clean_ballots, candidates=names), seats



from .ballot_generator import (  # noqa
    PlackettLuce,
    BradleyTerry,
    BallotSimplex,
    ImpartialCulture,
    ImpartialAnonymousCulture,
    CambridgeSampler,
    AlternatingCrossover,
)
from .election_types import (  # noqa
    STV,
    SNTV,
    SequentialRCV,
    Bloc,
    Borda,
    Limited,
    SNTV_STV_Hybrid,
    TopTwo,
    DominatingSets,
    CondoBorda,
    Pluarality,
)
from .ballot import Ballot  # noqa
from .pref_profile import PreferenceProfile  # noqa
from .cleaning import (  # noqa
    remove_empty_ballots,
    deduplicate_profiles,
    remove_noncands,
)
from .utils import (  # noqa
    compute_votes,
    fractional_transfer,
    random_transfer,
    seqRCV_transfer,
    first_place_votes,
    borda_scores,
)
from .cvr_loaders import rank_column_csv, blt  # noqa



from fractions import Fraction
from pydantic import BaseModel
from typing import Optional


class Ballot(BaseModel):
    """
    Data structure to represent a possible cast ballot.
    :param id: optional :class:`str` assigned ballot id
    :param ranking: :class:`list` of candidate ranking
    :param weight: :class:`float`/:class:`fraction` weight assigned to a given a ballot
    :param voters: optional :class:`list` of voters who cast a given a ballot
    """

    id: Optional[str] = None
    ranking: list[set]
    weight: Fraction
    voters: Optional[set[str]] = None

    class Config:
        arbitrary_types_allowed = True

    def __eq__(self, other):
        # Check type
        if not isinstance(other, Ballot):
            return False

        # Check id
        if self.id is not None:
            if self.id != other.id:
                return False

        # Check ranking
        if self.ranking != other.ranking:
            return False

        # Check weight
        if self.weight != other.weight:
            return False

        # Check voters
        if self.voters is not None:
            if self.voters != other.voters:
                return False

        return True

    def __hash__(self):
        return hash(str(self.ranking))



from abc import ABC, abstractmethod
from typing import Union, Any
from itertools import permutations
import math

from .ballot_generator import BallotGenerator
from .election_state import ElectionState
from .pref_profile import PreferenceProfile
from .ballot import Ballot


class Simulation(ABC):
    """
    Base class for model complex elections or statewide simulations as
    done in MGGG's RCV research.

    *Attributes*

        Ballots: PreferenceProfile or dictionary of ballot generators
    """

    def __init__(self, ballots: Union[PreferenceProfile, dict, None] = None):
        if ballots:
            self.ballots = ballots

    @abstractmethod
    def run_simulation(self) -> Any:
        """
        User written function to feed parameters into election simulation
        """
        pass

    @abstractmethod
    def sim_election(self) -> Union[ElectionState, list]:
        """
        Runs election(s) with specified parameters.
        """
        pass

    def generate_ballots(
        self, num_ballots: int, candidates: Union[list, dict], params: dict
    ) -> list[tuple[Any, PreferenceProfile]]:
        """
        Generates perference profiles if ballot generator models
        are assigned to the class.

        Can be overridden based on user needs.
        """
        if isinstance(self.ballots, PreferenceProfile):
            raise TypeError("No generator assigned to produce ballots")

        ballots = []
        for model_name, model in self.ballots.items():
            generator: BallotGenerator = model(
                candidates=candidates,
                hyperparams=params,
            )
            ballots.append((model_name, generator.generate_profile(num_ballots)))

        return ballots


class Election(ABC):
    """
    Abstract base class for election types.

    Includes functions to resolve input ties included in PreferenceProfile
    """

    def __init__(self, profile: PreferenceProfile, ballot_ties: bool = True):
        if ballot_ties:
            self._profile = self.resolve_input_ties(profile)
        else:
            self._profile = profile

        self.state = ElectionState(curr_round=0, profile=self._profile)

    @abstractmethod
    def run_step(self, *args: Any, **kwargs: Any):
        pass

    @abstractmethod
    def run_election(self, *args: Any, **kwargs: Any):
        pass

    def resolve_input_ties(self, profile: PreferenceProfile) -> PreferenceProfile:
        """
        Takes in a PeferenceProfile with potential ties in a ballot. Replaces
        ballots with ties with fractionally weighted ballots corresonding to
        all permutation of the tied ranking
        """
        new_ballots = []
        old_ballots = profile.get_ballots()

        for ballot in old_ballots:
            if not any(len(rank) > 1 for rank in ballot.ranking):
                new_ballots.append(ballot)
            else:
                num_ties = 0
                for rank in ballot.ranking:
                    if len(rank) > 1:
                        num_ties += 1

                resolved_ties = fix_ties(ballot)
                new_ballots += recursively_fix_ties(resolved_ties, num_ties)

        return PreferenceProfile(ballots=new_ballots)


# helpers
def recursively_fix_ties(ballot_lst: list[Ballot], num_ties: int) -> list[Ballot]:
    """
    Recursively fixes ties in a ballot in the case there is more then one tie
    """
    # base case, if only one tie to resolved return the list of already
    # resolved ballots
    if num_ties == 1:
        return ballot_lst

    # in the event multiple positions have ties
    else:
        update = set()
        for ballot in ballot_lst:
            update.update(set(fix_ties(ballot)))

        return recursively_fix_ties(list(update), num_ties - 1)


def fix_ties(ballot: Ballot) -> list[Ballot]:
    """
    Helper function for recursively_fix_ties. Resolves the first appearing
    tied rank in the input ballot by return list of permuted ballots
    """

    ballots = []
    for idx, rank in enumerate(ballot.ranking):
        if len(rank) > 1:
            for order in permutations(rank):
                resolved = []
                for cand in order:
                    resolved.append(set(cand))
                ballots.append(
                    Ballot(
                        id=ballot.id,
                        ranking=ballot.ranking[:idx]
                        + resolved
                        + ballot.ranking[idx + 1 :],
                        weight=ballot.weight / math.factorial(len(rank)),
                        voters=ballot.voters,
                    )
                )

    return ballots



from votekit.pref_profile import PreferenceProfile
from typing import Optional, Callable
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
from typing import Dict
from sklearn import manifold  # type: ignore

# Helper function for MDS Plot
def distance_matrix(
    pp_arr: list[PreferenceProfile], distance: Callable[..., int], *args, **kwargs
):
    """
    creates pairwise distance matrix between preference profiles. The i-th and j-th entry are
    pairwise distance between i-th col preference profile and the j-th row preference profile. \n
    pp_arr: list of preference profiles \n
    distance: callable distance function type. See distance.py
    """
    rows = len(pp_arr)
    dist_matrix = np.zeros((rows, rows))

    for i in range(rows):
        for j in range(i + 1, rows):
            dist_matrix[i][j] = distance(pp_arr[i], pp_arr[j], *args, **kwargs)
            dist_matrix[j][i] = dist_matrix[i][j]
    return dist_matrix


def plot_MDS(
    data: Dict[str, list[PreferenceProfile]],
    distance: Callable[..., int],
    marker_size: Optional[int] = 5,
    *args,
    **kwargs
):
    """
    Creates a multidimensional scaling plot \n
    data: dictionary with key being a 'color' and value being
    list of pp's. ex: {'color': list[PreferenceProfile]} \n
    distance: distance function \n
    """
    # combine all lists to create distance matrix
    combined_pp = []
    for pp_list in data.values():
        combined_pp.extend(pp_list)

    # compute distance matrix
    dist_matrix = distance_matrix(combined_pp, distance, *args, **kwargs)

    mds = manifold.MDS(
        n_components=2,
        max_iter=3000,
        eps=1e-9,
        dissimilarity="precomputed",
        n_jobs=1,
        normalized_stress="auto",
    )
    pos = mds.fit(np.array(dist_matrix)).embedding_

    # Plot and color data
    start_pos = 0
    for key, value_list in data.items():
        end_pos = start_pos + len(value_list)
        plt.scatter(
            pos[start_pos:end_pos, 0],
            pos[start_pos:end_pos, 1],
            color=key,
            lw=0,
            s=marker_size,
        )
        start_pos += len(value_list)
    plt.title("MDS Plot for Pair Wise Election Distances")
    plt.show()
    return plt



import pandas as pd
from pydantic import BaseModel
from typing import Optional

from .pref_profile import PreferenceProfile
from .utils import candidate_position_dict

pd.set_option("display.colheader_justify", "left")


class ElectionState(BaseModel):
    """
    Object that stores information on each round of a RCV election and the final outcome.
    :param curr_round: :class:`int` : current round number. Defaults to 0 before an election.
    :param elected: :class:`list[set[str]]` list of candidates who pass a threshold to win
    :param eliminated: :class:`list[set[str]]` list of candidates who were eliminated
    :param remaining: :class:`list[set[str]]` list of candidates who are still in the running
    :param rankings: :class: `list[set]` list ranking of candidates with sets representing ties
    :param profile: :class:`PreferenceProfile` an instance of a preference profile object
    :param previous: an instance of :class:`ElectionState` representing previous round
    """

    curr_round: int = 0
    elected: list[set[str]] = []
    eliminated: list[set[str]] = []
    remaining: list[set[str]] = []
    profile: PreferenceProfile
    previous: Optional["ElectionState"] = None

    class Config:
        allow_mutation = False

    def get_all_winners(self) -> list[set[str]]:
        """
        Returns a list of elected candidates ordered from first round to current round.
        :rtype: :class:`list[set[str]]`
        """
        if self.previous:
            return self.previous.get_all_winners() + self.elected
        else:
            return self.elected

    def get_all_eliminated(self) -> list[set[str]]:
        """
        Returns a list of eliminated candidates ordered from current round to first round
        :rtype: :class:`list[set[str]]`
        """
        if self.previous:
            return self.eliminated + self.previous.get_all_eliminated()
        else:
            return self.eliminated

    def get_rankings(self) -> list[set[str]]:
        """
        Returns list of all candidates in order of their ranking after each round
        :rtype: :class:`list[set[str]]`
        """
        if self.remaining != [{}]:
            return self.get_all_winners() + self.remaining + self.get_all_eliminated()
        else:
            return self.get_all_winners() + self.get_all_eliminated()

    def get_round_outcome(self, roundNum: int) -> dict:
        # {'elected':list[set[str]], 'eliminated':list[set[str]]}
        """
        returns a dictionary with elected and eliminated candidates
        :rtype: :class:`dict`
        """
        if self.curr_round == roundNum:
            return {
                "Elected": [c for s in self.elected for c in s],
                "Eliminated": [c for s in self.eliminated for c in s],
            }
        elif self.previous:
            return self.previous.get_round_outcome(roundNum)
        else:
            raise ValueError("Round number out of range")

    def changed_rankings(self) -> dict:
        """
        Returns dict of (key) string candidates who changed
        ranking from previous round and (value) a tuple of (previous rank, new rank)
        :rtype: :class:`dict`
        """

        if not self.previous:
            raise ValueError("This is the first round, cannot compare previous ranking")

        prev_ranking: dict = candidate_position_dict(self.previous.get_rankings())
        curr_ranking: dict = candidate_position_dict(self.get_rankings())
        if curr_ranking == prev_ranking:
            return {}

        changes = {}
        for candidate, index in curr_ranking.items():
            if prev_ranking[candidate] != index:
                changes[candidate] = (prev_ranking[candidate], index)
        return changes

    def status(self) -> pd.DataFrame:
        """
        Returns dataframe displaying candidate, status (elected, eliminated,
        remaining)
        :rtype: :class:`DataFrame`
        """
        all_cands = [c for s in self.get_rankings() for c in s]
        status_df = pd.DataFrame(
            {
                "Candidate": all_cands,
                "Status": ["Remaining"] * len(all_cands),
                "Round": [self.curr_round] * len(all_cands),
            }
        )

        for round in range(1, self.curr_round + 1):
            results = self.get_round_outcome(round)
            for status, candidates in results.items():
                for cand in candidates:
                    status_df.loc[status_df["Candidate"] == cand, "Status"] = status
                    status_df.loc[status_df["Candidate"] == cand, "Round"] = round

        return status_df

    def __str__(self):
        show = self.status()
        # show["Round"] = show["Round"].astype(str).str.rjust(3)
        # show["Status"] = show["Status"].str.ljust(10)
        return show.to_string(index=False, justify="justify")

    __repr__ = __str__

    ###############################################################################################

    # def add_winners_and_losers(self, winners: set[str], losers: set[str]) -> "Outcome":
    #   # example method, feel free to delete if not useful
    #  if not winners.issubset(self.remaining) or not losers.issubset(self.remaining):
    #     missing = (winners.difference(set(self.remaining)) |
    # (losersdifference(set(self.remaining)))
    #     raise ValueError(f"Cannot promote winners, {missing} not in remaining")
    # return Outcome(
    #     remaining=set(self.remaining).difference(winners | losers),
    #     elected=list(set(self.elected) | winners)
    #     eliminated=list(set(self.eliminated) | losers)
    # )

    #   def difference_remaining_candidates(
    #       self, prevOutcome1: "Outcome", prevOutcome2: "Outcome"
    #   ) -> float:
    #       """returns the fractional difference in number of
    #       remaining candidates; assumes ballots don't change by round
    #       """
    #       if (not prevOutcome1.get_profile()) or (not prevOutcome2.get_profile()):
    #           raise ValueError("Profile missing")
    # check if from same conshow
    #        elif set(prevOutcome1.get_profile().ballots) != set(
    #            prevOutcome2.get_profile().ballots
    #        ):
    #            raise ValueError("Cannot compare outcomes from different elections")
    #
    #        remaining_diff = len(
    #            (set(prevOutcome1.remaining)).difference(prevOutcome2.remaining)
    #        )
    #        allcandidates = len(prevOutcome1.get_profile().get_candidates())
    #        return remaining_diff / allcandidates



import csv
from fractions import Fraction
import pandas as pd
from pydantic import BaseModel, validator
from typing import Optional

from .ballot import Ballot


class PreferenceProfile(BaseModel):
    """
    ballots (list of ballots): ballots from an election
    candidates (list): list of candidates, can be user defined
    """

    ballots: list[Ballot] = []
    candidates: Optional[list] = None
    df: pd.DataFrame = pd.DataFrame()

    @validator("candidates")
    def cands_must_be_unique(cls, candidates: list) -> list:
        if not len(set(candidates)) == len(candidates):
            raise ValueError("all candidates must be unique")
        return candidates

    def get_ballots(self) -> list[Ballot]:
        """
        Returns list of ballots
        """
        return self.ballots

    # @cache
    def get_candidates(self) -> list:
        """
        Returns list of unique candidates
        """
        if self.candidates is not None:
            return self.candidates

        unique_cands: set = set()
        for ballot in self.ballots:
            unique_cands.update(*ballot.ranking)

        return list(unique_cands)

    # can also cache
    def num_ballots(self):
        """
        Assumes weights correspond to number of ballots given to a ranking
        """
        num_ballots = 0
        for ballot in self.ballots:
            num_ballots += ballot.weight

        return num_ballots

    def to_dict(self, standardize: bool) -> dict:
        """
        Converts ballots to dictionary with keys tuple(ranking) and values
        the corresponding total weights \n
        input: standardize to model an election distribution
        """
        num_ballots = self.num_ballots()
        di: dict = {}
        for ballot in self.ballots:
            rank_tuple = tuple(next(iter(item)) for item in ballot.ranking)
            if standardize:
                weight = ballot.weight / num_ballots
            else:
                weight = ballot.weight
            if rank_tuple not in di.keys():
                di[rank_tuple] = weight
            else:
                di[rank_tuple] += weight
        return di

    class Config:
        arbitrary_types_allowed = True

    def to_csv(self, fpath):
        """
        Saves Preference Profile to CSV
        Args:
            fpath (str): path to the saved csv
        """
        with open(fpath, "w", newline="") as csvfile:
            fieldnames = ["weight", "ranking"]
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            for ballot in self.ballots:
                writer.writerow({"weight": ballot.weight, "ranking": ballot.ranking})

    def create_df(self) -> pd.DataFrame:
        """
        Creates DF for display and building plots
        """
        weights = []
        ballots = []
        for ballot in self.ballots:
            part = []
            for ranking in ballot.ranking:
                for cand in ranking:
                    if len(ranking) > 2:
                        part.append(f"{cand} (Tie)")
                    else:
                        part.append(cand)
            ballots.append(tuple(part))
            weights.append(int(ballot.weight))

        df = pd.DataFrame({"Ballots": ballots, "Weight": weights})
        # df["Ballots"] = df["Ballots"].astype(str).str.ljust(60)
        df["Voter Share"] = df["Weight"] / df["Weight"].sum()
        # fill nans with zero for edge cases
        df["Voter Share"] = df["Voter Share"].fillna(0.0)
        # df["Weight"] = df["Weight"].astype(str).str.rjust(3)
        return df.reset_index(drop=True)

    def head(
        self, n: int, percents: Optional[bool] = False, totals: Optional[bool] = False
    ) -> pd.DataFrame:
        """
        Displays top-n ballots in profile based on weight
        """
        if self.df.empty:
            self.df = self.create_df()

        df = (
            self.df.sort_values(by="Weight", ascending=False)
            .head(n)
            .reset_index(drop=True)
        )

        if totals:
            df = sum_row(df)

        if not percents:
            return df.drop(columns="Voter Share")

        return df

    def tail(
        self, n: int, percents: Optional[bool] = False, totals: Optional[bool] = False
    ) -> pd.DataFrame:
        """
        Displays bottom-n ballots in profile based on weight
        """
        if self.df.empty:
            self.df = self.create_df()

        df = (
            self.df.sort_values(by="Weight", ascending=False)
            .head(n)
            .reset_index(drop=True)
        )
        if totals:
            df = sum_row(df)

        if not percents:
            return df.drop(columns="Voter Share")

        return df

    def __str__(self) -> str:
        """
        Displays top 15 or whole profiles
        """
        if self.df.empty:
            self.df = self.create_df()

        if len(self.df) < 15:
            return self.head(n=len(self.df)).to_string(index=False, justify="justify")

        return self.head(n=15).to_string(index=False, justify="justify")

    # set repr to print outputs
    __repr__ = __str__

    def condense_ballots(self):
        class_vector = []
        seen_rankings = []
        for ballot in self.ballots:
            if ballot.ranking not in seen_rankings:
                seen_rankings.append(ballot.ranking)
            class_vector.append(seen_rankings.index(ballot.ranking))

        new_ballot_list = []
        for i, ranking in enumerate(seen_rankings):
            total_weight = 0
            for j in range(len(class_vector)):
                if class_vector[j] == i:
                    total_weight += self.ballots[j].weight
            new_ballot_list.append(
                Ballot(ranking=ranking, weight=Fraction(total_weight))
            )
        self.ballots = new_ballot_list

    def __eq__(self, other):
        if not isinstance(other, PreferenceProfile):
            return False
        self.condense_ballots()
        other.condense_ballots()
        for b in self.ballots:
            if b not in other.ballots:
                return False
        for b in self.ballots:
            if b not in other.ballots:
                return False
        return True


def sum_row(df: pd.DataFrame) -> pd.DataFrame:
    """
    Computes sum total for weight and voter share column
    """
    sum_row = {
        "Ballot": "",
        "Weight": df["Weight"].sum(),
        "Voter Share": df["Voter Share"].sum(),
    }

    df.loc["Totals"] = sum_row  # type: ignore

    return df.fillna("")



from copy import deepcopy
from fractions import Fraction
from functools import reduce
from itertools import groupby
from typing import Callable

from .pref_profile import PreferenceProfile
from .ballot import Ballot


def remove_empty_ballots(
    pp: PreferenceProfile, keep_candidates: bool = False
) -> PreferenceProfile:
    """
    Removes empty ballots from a preference profile.
    :param pp: :class:`PreferenceProfile`
    :param keep_candidates: optional :class:`bool` If True, keep all of the
    candidates from the original preference profile in the returned preference
    profile, even if some no longer appear in the returned ballots.
    :return: A preference profile without empty ballots.
    :rtype: :class:`PreferenceProfile`
    """
    ballots_nonempty = [
        deepcopy(ballot) for ballot in pp.get_ballots() if ballot.ranking
    ]
    if keep_candidates:
        old_cands = deepcopy(pp.get_candidates())
        pp_clean = PreferenceProfile(ballots=ballots_nonempty, candidates=old_cands)
    else:
        pp_clean = PreferenceProfile(ballots=ballots_nonempty)
    return pp_clean


def _clean(
    pp: PreferenceProfile, clean_ballot_func: Callable[[Ballot], Ballot]
) -> PreferenceProfile:
    """
    General cleaning function that takes a preference profile and applies a
    cleaning function to each ballot and merges the ballots with the same ranking.
    This is used primarily when only the ballot ranking needs to be cleaned.
    :param pp: :class:`PreferenceProfile` The preference profile to be cleaned
    :param clean_ballot_func: :class:`Callable`[:class:`list`[:class:`Ballot`], :class:`Ballot`]
    Function that takes a list of ballots and cleans them
    :return: A cleaned preference profile
    :rtype: :class:`PreferenceProfile`
    """
    # apply cleaning function to clean all ballots
    if clean_ballot_func is not None:
        cleaned = map(clean_ballot_func, pp.ballots)
    # group ballots that have the same ranking after cleaning
    grouped_ballots = [
        list(result)
        for key, result in groupby(cleaned, key=lambda ballot: ballot.ranking)
    ]
    # merge ballots in the same groups
    new_ballots = [merge_ballots(b) for b in grouped_ballots]
    return PreferenceProfile(ballots=new_ballots)


def merge_ballots(ballots: list[Ballot]) -> Ballot:
    """
    Takes a list of ballots with the same ranking and merge them into one ballot.
    :param ballots: :class:`list`[:class:`Ballot`] Where each ballot has the same
    ranking.
    :return: A ballot with the same ranking and aggregated weight and voters
    :rtype: :class:`Ballot`
    """
    weight = sum(b.weight for b in ballots)
    ranking = ballots[0].ranking
    voters_to_merge = [b.voters for b in ballots if b.voters]
    voters = None
    if len(voters_to_merge) > 0:
        voters = reduce(lambda b1, b2: b1.union(b2), voters_to_merge)
        voters = set(voters)
    return Ballot(ranking=ranking, voters=voters, weight=Fraction(weight))


# TODO: Brenda will replace this function with the function she wrote,
# TODO: change to keep ranks so that we’ll have None
def deduplicate_profiles(pp: PreferenceProfile) -> PreferenceProfile:
    """
    Given a preference profile, deduplicates its ballots.
    :param pp: :class:`PreferenceProfile`
    :return: A preference profile without duplicates
    :rtype: :class:`PreferenceProfile`
    """

    def deduplicate_ballots(ballot: Ballot) -> Ballot:
        """
        Takes a ballot and deduplicates its rankings
        :param ballot: :class:`Ballot` with duplicates in its ranking
        :return: A ballot without duplicates
        :rtype: :class:`Ballot`
        """
        ranking = ballot.ranking
        dedup_ranking = []
        for cand in ranking:
            if cand in ranking and cand not in dedup_ranking:
                # dedup_ranking.append({None})
                dedup_ranking.append(cand)
        new_ballot = Ballot(
            id=ballot.id,
            weight=Fraction(ballot.weight),
            ranking=dedup_ranking,
            voters=ballot.voters,
        )
        return new_ballot

    pp_clean = _clean(pp=pp, clean_ballot_func=deduplicate_ballots)
    return pp_clean


def remove_from_ballots(ballot: Ballot, non_cands: list[str]) -> Ballot:
    """
    Removes non-candidiates from ballot objects
    """
    # TODO: adjust so string and list of strings are acceptable inputes

    to_remove = []
    for item in non_cands:
        to_remove.append({item})

    ranking = ballot.ranking
    clean_ranking = []
    for cand in ranking:
        if cand not in to_remove and cand not in clean_ranking:
            clean_ranking.append(cand)

    clean_ballot = Ballot(
        id=ballot.id,
        ranking=clean_ranking,
        weight=Fraction(ballot.weight),
        voters=ballot.voters,
    )

    return clean_ballot


## TODO: typehints so that non_cands inputs is str or list of strs
def remove_noncands(
    profile: PreferenceProfile, non_cands: list[str]
) -> PreferenceProfile:
    """
    Removes user-assigned non-candidates from ballots, deletes ballots
    that are empty as a result of the removal.
    :param profile: :class:`PreferenceProfile` Uncleaned preference profile.
    :param non_cands: :class:`list`[:class:`str`] Non-candidate items to be removed.
    :return: A profile with non-candidates removed.
    :rtype: :class:`PreferenceProfile`
    """
    cleaned = [
        remove_from_ballots(ballot, non_cands)
        for ballot in profile.ballots
        if remove_from_ballots(ballot, non_cands).ranking
    ]
    grouped_ballots = [
        list(result)
        for key, result in groupby(cleaned, key=lambda ballot: ballot.ranking)
    ]
    # merge ballots in the same groups
    new_ballots = [merge_ballots(b) for b in grouped_ballots]
    return PreferenceProfile(ballots=new_ballots)



from abc import ABC, abstractmethod
from functools import cache
import networkx as nx  # type: ignore
from typing import Callable, Optional, Union, Any

from ..pref_profile import PreferenceProfile
from ..utils import COLOR_LIST


class Graph(ABC):
    """
    Base class for graph models, contains general graph algorithmics
    applicable to any implementation
    """

    def __init__(self, graph: nx.Graph = None):
        self.graph = graph
        self.node_data: dict = {}  # store node to avoid acessing Nx.graph

    @abstractmethod
    def build_graph(self, *args: Any, **kwargs: Any) -> nx.Graph:
        pass

    @abstractmethod
    def draw(self, *args: Any, **kwags: Any):
        pass

    def distance_between_subsets(self, A: nx.Graph, B: nx.Graph):
        """Returns distance between A,B"""
        Gc = self.graph
        shortest_paths: list = []
        for a in A.nodes:
            shortest_paths.extend(nx.shortest_path_length(Gc, a, b) for b in B.nodes)
        return min(shortest_paths)

    def subgraph_neighborhood(self, center, radius: int = 2) -> nx.Graph:
        """Returns a ball around center of given radius in the graph of all ballots"""
        return nx.ego_graph(self.graph, center, radius)

    def k_heaviest_neighborhoods(self, k: int = 2, radius: int = 2):
        """Returns dict of k ball neighborhoods of
        given radius with their centers and weights
        """
        if not self.node_data or sum(self.node_data.values()) == 0:
            raise TypeError("no weights assigned to graph")

        cast_ballots = {x for x in self.node_data.keys() if self.node_data[x] > 0}

        max_balls = {}

        for _ in range(k):
            weight = 0
            if len(cast_ballots) == 0:
                break
            for center in cast_ballots:
                ball = self.subgraph_neighborhood(center, radius)
                relevant = cast_ballots.intersection(
                    set(ball.nodes)
                )  ##cast ballots inside the ball
                tmp = sum(self.node_data[node] for node in relevant)
                if tmp > weight:
                    weight = tmp
                    max_center = center
                    max_ball = ball

            not_cast_in_max_ball = set(max_ball.nodes).difference(cast_ballots)
            max_ball.remove_nodes_from(not_cast_in_max_ball)
            max_balls[max_center] = (max_ball, weight)

            cast_ballots = cast_ballots.difference(set(max_ball.nodes))

        return max_balls


class BallotGraph(Graph):
    """
    Class to build graphs for elections with possible incomplete ballots

    Inputs:
        source: (PreferenceProfile, number of candidates, or list of candiates),
                data to create graph from
        complete: (Optional[bool]) build complete graph or incomplete
    """

    def __init__(
        self,
        source: Union[PreferenceProfile, int, list],
        complete: Optional[bool] = True,
    ):
        super().__init__()

        self.profile = None
        self.candidates = None

        if isinstance(source, int):
            self.graph = self.build_graph(source)
            self.num_cands = source

        if isinstance(source, list):
            self.num_cands = len(source)
            self.graph = self.build_graph(len(source))
            self.candidates = source

        if isinstance(source, PreferenceProfile):
            self.profile = source
            self.num_voters = source.num_ballots()
            self.num_cands = len(source.get_candidates())
            if not self.graph:
                self.graph = self.build_graph(len(source.get_candidates()))
            self.graph = self.from_profile(source, complete)

        if not self.node_data:
            self.node_data = {ballot: 0 for ballot in self.graph.nodes}

        self.num_voters = sum(self.node_data.values())

    def _relabel(self, gr: nx.Graph, new_label: int, num_cands: int) -> nx.Graph:
        """Relabels nodes in gr based on new_label"""
        node_map = {}
        graph_nodes = list(gr.nodes)

        for k in graph_nodes:
            # add the value of new_label to every entry in every ballot
            tmp = [new_label + y for y in k]

            # reduce everything mod new_label
            for i in range(len(tmp)):
                if tmp[i] > num_cands:
                    tmp[i] = tmp[i] - num_cands
            node_map[k] = tuple([new_label] + tmp)

        return nx.relabel_nodes(gr, node_map)

    def build_graph(self, n: int) -> nx.Graph:  # ask Gabe about optimizing?
        """
        Builds graph of all possible ballots given a number of candiates
        """
        Gc = nx.Graph()
        # base cases
        if n == 1:
            Gc.add_nodes_from([(1)], weight=0, cast=False)

        elif n == 2:
            Gc.add_nodes_from([(1, 2), (2, 1)], weight=0, cast=False)
            Gc.add_edges_from([((1, 2), (2, 1))])

        elif n > 2:
            G_prev = self.build_graph(n - 1)
            for i in range(1, n + 1):
                # add the node for the bullet vote i
                Gc.add_node((i,), weight=0, cast=False)

                # make the subgraph for the ballots where i is ranked first
                G_corner = self._relabel(G_prev, i, n)

                # add the components from that graph to the larger graph
                Gc.add_nodes_from(G_corner.nodes, weight=0, cast=False)
                Gc.add_edges_from(G_corner.edges)

                # connect the bullet vote node to the appropriate vertices
                if n == 3:
                    Gc.add_edges_from([(k, (i,)) for k in G_corner.nodes])
                else:
                    Gc.add_edges_from(
                        [(k, (i,)) for k in G_corner.nodes if len(k) == 2]
                    )

            nodes = Gc.nodes

            new_edges = [
                (bal, (bal[1], bal[0]) + bal[2:]) for bal in nodes if len(bal) >= 2
            ]
            Gc.add_edges_from(new_edges)

        return Gc

    def from_profile(
        self, profile: PreferenceProfile, complete: Optional[bool] = True
    ) -> nx.Graph:
        """
        Updates existing graph based on cast ballots from a PreferenceProfile,
        or creates graph based on PreferenceProfile
        """
        if not self.profile:
            self.profile = profile

        if not self.num_voters:
            self.num_voters = profile.num_ballots()

        self.candidates = profile.get_candidates()
        ballots = profile.get_ballots()
        self.cand_num = self.number_cands(tuple(self.candidates))
        self.node_data = {ballot: 0 for ballot in self.graph.nodes}

        for ballot in ballots:
            ballot_node = []
            for position in ballot.ranking:
                if len(position) > 1:
                    raise ValueError(
                        "ballots must be cleaned to resolve ties"
                    )  # still unsure about ties
                for cand in position:
                    ballot_node.append(self.cand_num[cand])
            if len(ballot_node) == len(self.candidates) - 1:
                ballot_node = self.fix_short_ballot(
                    ballot_node, list(self.cand_num.values())
                )

            if tuple(ballot_node) in self.graph.nodes:
                self.graph.nodes[tuple(ballot_node)]["weight"] += ballot.weight
                self.graph.nodes[tuple(ballot_node)]["cast"] = True
                self.node_data[tuple(ballot_node)] += ballot.weight

        if not complete:
            partial = nx.Graph()
            for node in self.graph.nodes:
                if self.graph.nodes[node]["cast"]:
                    partial.add_node(
                        node,
                        weight=self.graph.nodes[node]["weight"],
                        cast=self.graph.nodes[node]["cast"],
                    )

            self.graph = partial

        return self.graph

    def fix_short_ballot(self, ballot: list, candidates: list) -> list:
        """
        Appends short ballots of n-1 length to add to BallotGraph
        """
        missing = set(candidates).difference(set(ballot))

        return ballot + list(missing)

    def label_cands(self, candidates):
        """
        Assigns candidate labels to ballot graph for plotting.

        Returns: a dictionary with keys, the index  of a node and values, the
        ballot with candidate names.
        """

        candidate_numbers = self.number_cands(tuple(candidates))

        cand_dict = {value: key for key, value in candidate_numbers.items()}

        cand_labels = {}
        for node in self.graph.nodes:
            ballot = []
            for num in node:
                ballot.append(cand_dict[num])
            cand_labels[node] = tuple(ballot)

        return cand_labels

    @cache
    def number_cands(self, cands: tuple) -> dict:
        """
        Assigns numerical marker to candidates
        """
        legend = {}
        for idx, cand in enumerate(cands):
            legend[cand] = idx + 1

        return legend

    def draw(self, neighborhoods: Optional[dict] = {}, labels: Optional[bool] = False):
        """visualize the whole election or select neighborhoods in the election."""
        # TODO: change this so that neighborhoods can have any neighborhood
        # not just heavy balls, also there's something wrong with the shades
        Gc = self.graph
        GREY = (0.44, 0.5, 0.56)
        node_cols: list = []
        node_labels = None

        k = len(neighborhoods) if neighborhoods else self.num_cands
        if k > len(COLOR_LIST):
            raise ValueError("Number of neighborhoods exceeds colors for plotting")
        cols = COLOR_LIST[:k]

        # self._clean()
        for ballot in Gc.nodes:
            i = -1
            color: tuple = GREY

            if neighborhoods:
                for center, neighborhood in neighborhoods.items():
                    neighbors, _ = neighborhood
                    if ballot in neighbors:
                        i = (list(neighborhoods.keys())).index(center)
                        break
            elif self.node_data[ballot] != 0 and self.profile:
                print(ballot)
                i = (list(self.cand_num.values())).index(ballot[0])

            if "weight" in ballot:
                color = tuple(ballot.weight * x for x in cols[i])
            node_cols.append(color)

        if labels:
            if not self.candidates:
                raise ValueError("no candidate names assigned")
            if self.candidates:
                node_labels = self.label_cands(self.candidates)
            elif self.profile:
                node_labels = self.label_cands(self.profile.get_candidates())

        nx.draw_networkx(Gc, with_labels=True, node_color=node_cols, labels=node_labels)

    # what are these functions supposed to do?
    def compare(self, new_pref: PreferenceProfile, dist_type: Callable):
        """compares the ballots of current and new profile"""
        raise NotImplementedError("Not yet built")

    def compare_rcv_results(self, new_pref: PreferenceProfile):
        """compares election results of current and new profle"""
        raise NotImplementedError("Not yet built")



from fractions import Fraction
import itertools as it
import numpy as np
from typing import Callable, Optional

from .models import Election
from .election_state import ElectionState
from .graphs.pairwise_comparison_graph import PairwiseComparisonGraph
from .pref_profile import PreferenceProfile
from .utils import (
    compute_votes,
    remove_cand,
    fractional_transfer,
    borda_scores,
    seqRCV_transfer,
    scores_into_set_list,
    tie_broken_ranking,
    elect_cands_from_set_ranking,
    first_place_votes,
)


class STV(Election):
    """
    Class for single-winner IRV and multi-winner STV elections

     **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `transfer`
    :   transfer method (e.g. fractional transfer)

    `seats`
    :   number of seats to be elected

    `quota`
    :   formula to calculate quota (defaults to droop)

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        transfer: Callable,
        seats: int,
        quota: str = "droop",
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        # let parent class handle the og profile and election state
        super().__init__(profile, ballot_ties)

        self.transfer = transfer
        self.seats = seats
        self.tiebreak = tiebreak
        self.threshold = self.get_threshold(quota)

    # can cache since it will not change throughout rounds
    def get_threshold(self, quota: str) -> int:
        """
        Calculates threshold required for election

        Args:
            quota: Type of quota formula

        Returns:
            Value of the threshold
        """
        quota = quota.lower()
        if quota == "droop":
            return int(self._profile.num_ballots() / (self.seats + 1) + 1)
        elif quota == "hare":
            return int(self._profile.num_ballots() / self.seats)
        else:
            raise ValueError("Misspelled or unknown quota type")

    def next_round(self) -> bool:
        """
        Determines if the number of seats has been met to call an election

        Returns:
            True if number of seats has been met, False otherwise
        """
        cands_elected = 0
        for s in self.state.get_all_winners():
            cands_elected += len(s)
        return cands_elected < self.seats

    def run_step(self) -> ElectionState:
        """
        Simulates one round an STV election

        Returns:
           An ElectionState object for a given round
        """
        remaining = self.state.profile.get_candidates()
        ballots = self.state.profile.get_ballots()
        round_votes = compute_votes(remaining, ballots)
        elected = []
        eliminated = []

        # if number of remaining candidates equals number of remaining seats,
        # everyone is elected
        if len(remaining) == self.seats - len(self.state.get_all_winners()):
            elected = [{cand} for cand, votes in round_votes]
            remaining = []
            ballots = []

        # elect all candidates who crossed threshold
        elif round_votes[0].votes >= self.threshold:
            for candidate, votes in round_votes:
                if votes >= self.threshold:
                    elected.append({candidate})
                    remaining.remove(candidate)
                    ballots = self.transfer(
                        candidate,
                        ballots,
                        {cand: votes for cand, votes in round_votes},
                        self.threshold,
                    )
        # since no one has crossed threshold, eliminate one of the people
        # with least first place votes
        elif self.next_round():
            lp_candidates = [
                candidate
                for candidate, votes in round_votes
                if votes == round_votes[-1].votes
            ]

            lp_cand = tie_broken_ranking(
                ranking=[set(lp_candidates)],
                profile=self.state.profile,
                tiebreak=self.tiebreak,
            )[-1]
            eliminated.append(lp_cand)
            ballots = remove_cand(lp_cand, ballots)
            remaining.remove(next(iter(lp_cand)))

        if len(elected) >= 1:
            elected = scores_into_set_list(
                first_place_votes(self.state.profile), [c for s in elected for c in s]
            )

        # Make sure list-of-sets have non-empty elements
        elected = [s for s in elected if s != set()]
        eliminated = [s for s in eliminated if s != set()]

        remaining = [set(remaining)]
        remaining = [s for s in remaining if s != set()]

        self.state = ElectionState(
            curr_round=self.state.curr_round + 1,
            elected=elected,
            eliminated=eliminated,
            remaining=remaining,
            profile=PreferenceProfile(ballots=ballots),
            previous=self.state,
        )
        return self.state

    def run_election(self) -> ElectionState:
        """
        Runs complete STV election

        Returns:
            An ElectionState object with results for a complete election
        """
        if not self.next_round():
            raise ValueError(
                f"Length of elected set equal to number of seats ({self.seats})"
            )

        while self.next_round():
            self.run_step()

        return self.state


class Limited(Election):
    """
    Limited: Elects m (seats) candidates with the highest k-approval scores.
    The k-approval score of a candidate is equal to the number of voters who \n
    rank this candidate among their k top ranked candidates.

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `k`
    :   value of an approval score

    `seats`
    :   number of seats to be elected

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        k: int,
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.k = k
        self.tiebreak = tiebreak

    def run_step(self) -> ElectionState:
        """
        Conducts Limited election in which m-candidates are elected based
        on approval scores

        Returns:
           An ElectionState object for a Limited election
        """
        profile = self.state.profile
        candidates = profile.get_candidates()
        candidate_approvals = {c: Fraction(0) for c in candidates}

        for ballot in profile.get_ballots():
            # First we have to determine which candidates are approved
            # i.e. in first k ranks on a ballot
            approvals = []
            for i, cand_set in enumerate(ballot.ranking):
                # If list of total candidates before and including current set
                # are less than seat count, all candidates are approved
                if len(list(it.chain(*ballot.ranking[: i + 1]))) < self.k:
                    approvals.extend(list(cand_set))
                # If list of total candidates before current set
                # are greater than seat count, no candidates are approved
                elif len(list(it.chain(*ballot.ranking[:i]))) > self.k:
                    approvals.extend([])
                # Else we know the cutoff is in the set, we compute and randomly
                # select the number of candidates we can select
                else:
                    accepted = len(list(it.chain(*ballot.ranking[:i])))
                    num_to_allow = self.k - accepted
                    approvals.extend(
                        np.random.choice(list(cand_set), num_to_allow, replace=False)
                    )

            # Add approval votes equal to ballot weight (i.e. number of voters with this ballot)
            for cand in approvals:
                candidate_approvals[cand] += ballot.weight

        # Order candidates by number of approval votes received
        ranking = scores_into_set_list(candidate_approvals)
        ranking = tie_broken_ranking(
            ranking=ranking, profile=self.state.profile, tiebreak=self.tiebreak
        )
        elected, eliminated = elect_cands_from_set_ranking(
            ranking=ranking, seats=self.seats
        )
        new_state = ElectionState(
            curr_round=self.state.curr_round + 1,
            elected=elected,
            eliminated=eliminated,
            remaining=list(),
            profile=PreferenceProfile(),
            previous=self.state,
        )
        self.state = new_state
        return self.state

    def run_election(self) -> ElectionState:
        """
        Simulates a complete Limited election

        Returns:
            An ElectionState object with results for a complete election
        """
        self.run_step()
        return self.state


class Bloc(Election):
    """
    Bloc: Elects m candidates with the highest m-approval scores. The m-approval \n
    score of a candidate is equal to the number of voters who rank this \n
    candidate among their m top ranked candidates.

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `seats`
    :   number of seats to be elected

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.tiebreak = tiebreak

    def run_step(self) -> ElectionState:
        """
        Conducts a Limited election to elect m-candidates

        Returns:
           An ElectionState object for a Limited election
        """
        limited_equivalent = Limited(
            profile=self.state.profile,
            seats=self.seats,
            k=self.seats,
            tiebreak=self.tiebreak,
        )
        outcome = limited_equivalent.run_election()
        self.state = outcome
        return outcome

    def run_election(self) -> ElectionState:
        """
        Runs complete Bloc election

        Returns:
            An ElectionState object with results for a complete election
        """
        self.run_step()
        return self.state


class SNTV(Election):
    """
    Single nontransferable vote (SNTV): Elects k-candidates with the highest \n
    Plurality scores

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `seats`
    :   number of seats to be elected

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.tiebreak = tiebreak

    def run_step(self) -> ElectionState:
        """
        Conducts a Limited election to elect k-candidates

        Returns:
           An ElectionState object for a Limited election
        """
        limited_equivalent = Limited(
            profile=self.state.profile, seats=self.seats, k=1, tiebreak=self.tiebreak
        )
        outcome = limited_equivalent.run_election()
        self.state = outcome
        return outcome

    def run_election(self) -> ElectionState:
        """
        Runs complete SNTV election

        Returns:
            An ElectionState object with results for a complete election
        """
        self.run_step()
        return self.state


class SNTV_STV_Hybrid(Election):
    """
    SNTV-IRV Hybrid: This method first runs SNTV to a cutoff, then
    runs STV to pick a committee with a given number of seats.

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `transfer`
    :   transfer method (e.g. fractional transfer)

    `r1_cutoff`
    :   first-round cutoff value

    `seats`
    :   number of seats to be elected

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        transfer: Callable,
        r1_cutoff: int,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.transfer = transfer
        self.r1_cutoff = r1_cutoff
        self.seats = seats
        self.tiebreak = tiebreak
        self.stage = "SNTV"  # SNTV, switches to STV, then Complete

    def run_step(self, stage: str) -> ElectionState:
        """
        Simulates one round an SNTV_STV election

        Args:
            stage: Stage of the hybrid election, can be SNTV or STV

        Returns:
           An ElectionState object for a given round
        """
        profile = self.state.profile

        new_state = None
        if stage == "SNTV":
            round_state = SNTV(
                profile=profile, seats=self.r1_cutoff, tiebreak=self.tiebreak
            ).run_election()

            # The STV election will be run on the new election state
            # Therefore we should not add any winners, but rather
            # set the SNTV winners as remaining candidates and update pref profiles
            new_profile = PreferenceProfile(
                ballots=remove_cand(
                    set().union(*round_state.eliminated), profile.get_ballots()
                )
            )
            new_state = ElectionState(
                curr_round=self.state.curr_round + 1,
                elected=list(),
                eliminated=round_state.eliminated,
                remaining=[set(new_profile.get_candidates())],
                profile=new_profile,
                previous=self.state,
            )
        elif stage == "STV":
            round_state = STV(
                profile=profile,
                transfer=self.transfer,
                seats=self.seats,
                tiebreak=self.tiebreak,
            ).run_election()

            new_state = ElectionState(
                curr_round=self.state.curr_round + 1,
                elected=round_state.get_all_winners(),
                eliminated=round_state.get_all_eliminated(),
                remaining=round_state.remaining,
                profile=round_state.profile,
                previous=self.state,
            )

        # Update election stage to cue next run step
        if stage == "SNTV":
            self.stage = "STV"
        elif stage == "STV":
            self.stage = "Complete"

        self.state = new_state  # type: ignore
        return new_state  # type: ignore

    def run_election(self) -> ElectionState:
        """
        Runs complete SNTV_STV election

        Returns:
            An ElectionState object with results for a complete election
        """
        while self.stage != "Complete":
            self.run_step(self.stage)
        return self.state  # type: ignore


class TopTwo(Election):
    """
    Top Two: Eliminates all but the top two plurality vote getters,and then
    conducts a runoff between them, reallocating other ballots

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `seats`
    :   number of seats to be elected

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.tiebreak = tiebreak

    def run_step(self) -> ElectionState:
        """
        Conducts a hybrid election for one seat with a cutoff of 2 for the runoff

        Returns:
            An ElectionState object for the hybrid election
        """
        hybrid_equivalent = SNTV_STV_Hybrid(
            profile=self.state.profile,
            transfer=fractional_transfer,
            r1_cutoff=2,
            seats=1,
            tiebreak=self.tiebreak,
        )
        outcome = hybrid_equivalent.run_election()
        self.state = outcome
        return outcome

    def run_election(self) -> ElectionState:
        """
        Simulates a complete TopTwo election

        Returns:
            An ElectionState object for a complete election
        """
        self.run_step()
        return self.state


class DominatingSets(Election):
    """
    Finds tiers of candidates by dominating set,which is a set of candidates
    such that every candidate in the set wins head to head comparisons against
    candidates outside of it

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties


    **Methods**
    """

    def __init__(self, profile: PreferenceProfile, ballot_ties: bool = True):
        super().__init__(profile, ballot_ties)

    def run_step(self) -> ElectionState:
        """
        Conducts a complete DominatingSets election as it is not a round-by-round
        system

        Returns:
            An ElectionState object for a complete election
        """
        pwc_graph = PairwiseComparisonGraph(self.state.profile)
        dominating_tiers = pwc_graph.dominating_tiers()
        if len(dominating_tiers) == 1:
            new_state = ElectionState(
                curr_round=self.state.curr_round + 1,
                elected=list(),
                eliminated=dominating_tiers,
                remaining=list(),
                profile=PreferenceProfile(),
                previous=self.state,
            )
        else:
            new_state = ElectionState(
                curr_round=self.state.curr_round + 1,
                elected=[set(dominating_tiers[0])],
                eliminated=dominating_tiers[1:],
                remaining=list(),
                profile=PreferenceProfile(),
                previous=self.state,
            )
        self.state = new_state
        return new_state

    def run_election(self) -> ElectionState:
        """
        Simulates a complete DominatingSets election

        Returns:
            An ElectionState object for a complete election
        """
        self.run_step()
        return self.state


class CondoBorda(Election):
    """
    Condo-Borda: Elects candidates ordered by dominating set, but breaks ties
    between candidates with Borda

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `seats`
    :   number of seats to be elected

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.tiebreak = tiebreak

    def run_step(self) -> ElectionState:
        """
        Conducts a complete Conda-Borda election as it is not a round-by-round
        system

        Returns:
            An ElectionState object for a complete election
        """
        pwc_graph = PairwiseComparisonGraph(self.state.profile)
        dominating_tiers = pwc_graph.dominating_tiers()
        ranking = tie_broken_ranking(
            ranking=dominating_tiers, profile=self.state.profile, tiebreak="borda"
        )
        elected, eliminated = elect_cands_from_set_ranking(
            ranking=ranking, seats=self.seats
        )

        new_state = ElectionState(
            curr_round=self.state.curr_round + 1,
            elected=elected,
            eliminated=eliminated,
            remaining=list(),
            profile=PreferenceProfile(),
            previous=self.state,
        )
        self.state = new_state
        return new_state

    def run_election(self) -> ElectionState:
        """
        Simulates a complete Conda-Borda election

        Returns:
            An ElectionState object for a complete election
        """
        self.run_step()
        return self.state


class SequentialRCV(Election):
    """
    Class to conduct Sequential RCV election, in which votes are not transferred
    after a candidate has reached threshold, or been elected

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `seats`
    :   number of seats to be elected

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.tiebreak = tiebreak

    def run_step(self, old_profile: PreferenceProfile) -> ElectionState:
        """
        Simulates a single step of the sequential RCV contest or a full
        IRV election run on the current set of candidates

         Returns:
           An ElectionState object for a given round
        """
        old_election_state = self.state

        IRVrun = STV(
            old_profile, transfer=seqRCV_transfer, seats=1, tiebreak=self.tiebreak
        )
        old_election = IRVrun.run_election()
        elected_cand = old_election.get_all_winners()[0]

        # Removes elected candidate from Ballot List
        updated_ballots = remove_cand(elected_cand, old_profile.get_ballots())

        # Updates profile with removed candidates
        updated_profile = PreferenceProfile(ballots=updated_ballots)

        self.state = ElectionState(
            curr_round=old_election_state.curr_round + 1,
            elected=[elected_cand],
            profile=updated_profile,
            previous=old_election_state,
            remaining=old_election.remaining,
        )
        return self.state

    def run_election(self) -> ElectionState:
        """
        Simulates a complete sequential RCV contest.

        Returns:
            An ElectionState object for a complete election
        """
        old_profile = self._profile
        elected = []  # type: ignore
        seqRCV_step = self.state

        while len(elected) < self.seats:
            seqRCV_step = self.run_step(old_profile)
            elected.append(seqRCV_step.elected)
            old_profile = seqRCV_step.profile
        return seqRCV_step


class Borda(Election):
    """
    Positional voting system that assigns a decreasing number of points to
    candidates based on order and a score vector. The conventional score
    vector is linear (n, n-1, ... 1)

    **Attributes**

    `profile`
    :   PreferenceProfile to run election on

    `seats`
    :   number of seats to be elected

    `score_vector`
    :   (Optional) weights assigned to candidate ranking

    `ballot_ties`
    :   (Optional) resolves input ballot ties if True, else assumes ballots have no ties

    'tiebreak'
    :   (Optional) resolves procedural and final ties by specified tiebreak

    **Methods**
    """

    def __init__(
        self,
        profile: PreferenceProfile,
        seats: int,
        score_vector: Optional[list[Fraction]],
        ballot_ties: bool = True,
        tiebreak: str = "random",
    ):
        super().__init__(profile, ballot_ties)
        self.seats = seats
        self.tiebreak = tiebreak
        self.score_vector = score_vector

    def run_step(self) -> ElectionState:
        """
        Simulates a complete Borda contest as Borda is not a round-by-round
        system

        Returns:
            An ElectionState object for a complete election
        """
        borda_dict = borda_scores(
            profile=self.state.profile, score_vector=self.score_vector
        )

        ranking = scores_into_set_list(borda_dict)
        ranking = tie_broken_ranking(
            ranking=ranking, profile=self.state.profile, tiebreak=self.tiebreak
        )

        elected, eliminated = elect_cands_from_set_ranking(
            ranking=ranking, seats=self.seats
        )

        new_state = ElectionState(
            curr_round=self.state.curr_round + 1,
            elected=elected,
            eliminated=eliminated,
            remaining=list(),
            profile=PreferenceProfile(),
            previous=self.state,
        )
        self.state = new_state
        return new_state

    def run_election(self) -> ElectionState:
        """
        Simulates a complete Borda contest

        Returns:
            An ElectionState object for a complete election
        """
        self.run_step()
        return self.state


class Pluarality(SNTV):
    """
    Simulates a single or multi-winner plurality election. Inherits
    methods from SNTV to run election
    """



import matplotlib.pyplot as plt  # type: ignore
from ..pref_profile import PreferenceProfile
from ..utils import first_place_votes, mentions, COLOR_LIST, borda_scores
from matplotlib.figure import Figure  # type: ignore


def plot_summary_stats(
    profile: PreferenceProfile, stat: str, multi_color: bool = True
) -> Figure:
    """
    Plots histogram of election summary statistics
    """
    stats = {
        "first place votes": first_place_votes,
        "mentions": mentions,
        "borda": borda_scores,
    }

    stat_func = stats[stat]
    data: dict = stat_func(profile)  # type: ignore

    if stat == "first place votes":
        ylabel = "First Place Votes"
    elif stat == "mentions":
        ylabel = "Total Mentions"
    else:
        ylabel = "Borda Scores"

    if multi_color:
        colors = COLOR_LIST[: len(list(data.keys()))]
    else:
        colors = [COLOR_LIST[-1]]

    fig, ax = plt.subplots()

    ax.bar(data.keys(), data.values(), color=colors, width=0.35)
    ax.set_xlabel("Candidates")
    ax.set_ylabel(ylabel)

    return fig


def accumlation_chart():
    """
    TODO: After 0.0.0 version release
    """
